# --- START OF MODIFIED FILE python_mcp_server.py ---

import os
import base64
import logging
import tempfile
import zipfile
import xml.etree.ElementTree as ET
from xml.dom import minidom
import re
import time
import json
from typing import List, Dict, Any, Optional
import random # Added for generate_test_data
import datetime # Added for generate_test_data

# Attempt to import Google libraries, handle if not installed
try:
    from googleapiclient.discovery import build
    from googleapiclient.errors import HttpError
    GOOGLE_LIBS_AVAILABLE = True
except ImportError:
    GOOGLE_LIBS_AVAILABLE = False
    HttpError = Exception # Define HttpError as base Exception if library missing

import requests
from dotenv import load_dotenv
from simple_salesforce import Salesforce
from mcp.server.fastmcp import FastMCP

# --- Configuration ---

# Configure logging
logging.basicConfig(
    level=logging.INFO,
    format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
)
logger = logging.getLogger("salesforce-mcp")

# Load environment variables from .env file if it exists
load_dotenv()

# API Versions (Update as needed)
SALESFORCE_API_VERSION = "59.0"  # Use a recent, supported API version

# Create an MCP server
mcp = FastMCP(
    "Salesforce Architect Assistant",
    # MODIFIED: Updated description to reflect automatic connection
    description="Provides tools to interact with Salesforce metadata and data. Automatically connects using SF_USERNAME, SF_PASSWORD, and SF_SECURITY_TOKEN environment variables when a tool requires Salesforce access.",
)

# --- Global State ---

# Global Salesforce connection object, initialized to None
sf: Optional[Salesforce] = None

# --- Helper Functions ---

# Ensure helper functions are defined before they are called by tools

def to_api_name(name: str, is_object: bool = False) -> str:
    """
    Converts a human-readable name to a valid Salesforce API name (e.g., My_Object__c).

    Args:
        name: The human-readable name.
        is_object: True if converting an object name, False for a field name.

    Returns:
        A valid Salesforce API name string.
    """
    # Replace invalid characters with underscores
    api_name = re.sub(r"[^a-zA-Z0-9_]", "_", name)
    # Replace multiple consecutive underscores with a single one
    api_name = re.sub(r"__+", "_", api_name)
    # Remove leading/trailing underscores
    api_name = api_name.strip("_")

    # Ensure it starts with a letter
    if not api_name or not api_name[0].isalpha():
        api_name = "X_" + api_name

    # Salesforce API Name Length Limit (40 for custom objects/fields)
    max_length = 40 - 3 # Account for __c suffix
    api_name = api_name[:max_length]
    api_name = api_name.strip('_') # Reclean after truncation

    # Append __c suffix
    if not api_name.endswith("__c"):
        api_name += "__c"

    # Convention: Uppercase first letter for objects
    if is_object and len(api_name) > 3: # Check length to avoid index error on just "__c"
         api_name = api_name[0].upper() + api_name[1:]

    return api_name

def get_sf_instance_url() -> Optional[str]:
    """Safely get the Salesforce instance URL from the connection object."""
    if sf and hasattr(sf, 'sf_instance'):
        return sf.sf_instance
    return None

def get_sf_session_id() -> Optional[str]:
    """Safely get the Salesforce session ID from the connection object."""
    if sf and hasattr(sf, 'session_id'):
        return sf.session_id
    return None

# NEW: Internal function to handle connection attempt using environment variables
# NEW: Internal function to handle connection attempt using environment variables
def _attempt_auto_connect() -> Optional[Salesforce]:
    """
    Attempts to establish a Salesforce connection using environment variables.
    Reads SF_USERNAME, SF_PASSWORD, SF_SECURITY_TOKEN.

    Returns:
        A Salesforce instance on success, None on failure (missing vars or connection error).
    """
    # Explicitly try loading dotenv again here just in case, although top-level should be sufficient
    # You could remove this later if the top-level load proves reliable.
    load_dotenv(verbose=True) # verbose=True logs if it finds and loads a .env file

    logger.info("Attempting automatic Salesforce connection using environment variables...")
    username = os.getenv("SF_USERNAME")
    password = os.getenv("SF_PASSWORD")
    security_token = os.getenv("SF_SECURITY_TOKEN", "") # Default to empty string if missing

    # --- DETAILED LOGGING ---
    logger.info(f"Read SF_USERNAME: {'Set (length ' + str(len(username)) + ')' if username else '*** Not Set or Empty ***'}")
    logger.info(f"Read SF_PASSWORD: {'Set' if password else '*** Not Set or Empty ***'}") # Avoid logging length/value
    logger.info(f"Read SF_SECURITY_TOKEN: {'Set (length ' + str(len(security_token)) + ')' if security_token else 'Not Set or Empty (optional)'}")
    # --- END DETAILED LOGGING ---

    if not username or not password:
        logger.error("SF_USERNAME and SF_PASSWORD environment variables are required and were not found or are empty.")
        return None

    # Combine password and token - standard simple-salesforce approach
    effective_password = password + security_token

    try:
        # Directly attempt connection using the combined password+token
        logger.info(f"Connecting to Salesforce API version {SALESFORCE_API_VERSION}...")
        sf_instance = Salesforce(
            username=username,
            password=password,            # Just the password from env var
            security_token=security_token, # Pass the token from env var explicitly
            version=SALESFORCE_API_VERSION
        )
        logger.info(f"Initial Salesforce() call successful. Instance: {sf_instance.sf_instance}. Verifying connection...")

        # Perform a simple query to verify the connection is truly established and user has basic permissions
        org_info = sf_instance.query("SELECT Id, Name FROM Organization LIMIT 1")
        if org_info and org_info.get("records"):
            org_name = org_info["records"][0].get("Name")
            logger.info(f"Successfully auto-connected and verified with Salesforce org: {org_name} on instance {sf_instance.sf_instance}")
            return sf_instance
        else:
             logger.error("Auto-connection verification query failed. Salesforce() call succeeded, but could not query Organization details. Check user permissions or org state.")
             return None # Treat as failure if basic query fails

    except Exception as e:
        # Use logger.exception to include the full traceback in the logs
        logger.exception(f"Automatic Salesforce connection failed during Salesforce() call or query verification.")
        # Provide more specific feedback based on common errors
        error_message = str(e)
        user_friendly_error = error_message # Default to original error
        if "INVALID_LOGIN" in error_message or "LOGIN_MUST_USE_SECURITY_TOKEN" in error_message:
            user_friendly_error = "Login failed: Invalid credentials or missing/incorrect security token. Check SF_USERNAME, SF_PASSWORD, SF_SECURITY_TOKEN."
        elif "Could not find a match for connection" in error_message:
             user_friendly_error = "Connection failed: Could not resolve Salesforce instance. Check username format (e.g., include .sandbox suffix if needed), credentials, and network connectivity."
        elif "timeout" in error_message.lower():
             user_friendly_error = "Connection failed: Request timed out. Check network connectivity to Salesforce."
        elif "refused" in error_message.lower():
             user_friendly_error = "Connection failed: Connection refused. Check network/firewall settings."

        logger.error(f"User-facing connection error hint: {user_friendly_error}")
        return None # Return None on any connection failure

# MODIFIED: Updated check_connection to attempt auto-connect if needed
def check_connection() -> bool:
    """
    Check if the Salesforce connection is established.
    If not, attempts to connect automatically using environment variables.

    Returns:
        True if connected, False otherwise.
    """
    global sf
    if sf:
        # Optional: Add a quick ping here if desired, but get_salesforce_connection_status does this
        # try:
        #    sf.query("SELECT Id FROM Organization LIMIT 1")
        #    return True
        # except Exception:
        #    logger.warning("Existing Salesforce connection seems invalid. Resetting.")
        #    sf = None
        #    return False # Force re-attempt if ping fails
        return True # Assume existing sf object is valid for now

    # If sf is None, attempt to connect
    logger.info("Salesforce connection not found, attempting automatic connection...")
    sf_instance = _attempt_auto_connect()
    if sf_instance:
        sf = sf_instance # Assign to global variable
        return True
    else:
        # Connection attempt failed (error already logged by _attempt_auto_connect)
        sf = None # Ensure sf is None if connection failed
        logger.warning("Automatic Salesforce connection failed. Check environment variables and logs.")
        return False


def prettify_xml(elem: ET.Element) -> str:
    """Return a pretty-printed XML string for the Element."""
    rough_string = ET.tostring(elem, "utf-8")
    reparsed = minidom.parseString(rough_string)
    # Indent with 4 spaces for typical package.xml format
    return reparsed.toprettyxml(indent="    ")

def check_deployment_status(instance_url: str, session_id: str, deploy_id: str) -> Dict[str, Any]:
    """
    Check the status of a Metadata API deployment using SOAP.

    Args:
        instance_url: The Salesforce instance URL (e.g., 'yourdomain.my.salesforce.com').
        session_id: The active Salesforce session ID.
        deploy_id: The asynchronous deployment process ID.

    Returns:
        A dictionary containing the deployment status ('success': bool, 'message'/'error', 'details').
    """
    if not instance_url or not session_id:
        return {"success": False, "error": "Missing instance URL or session ID for polling."}

    try:
        soap_url = f"https://{instance_url}/services/Soap/m/{SALESFORCE_API_VERSION}"
        metadata_ns = "http://soap.sforce.com/2006/04/metadata"
        soapenv_ns = "http://schemas.xmlsoap.org/soap/envelope/"
        ns_map = {"met": metadata_ns, "soapenv": soapenv_ns}

        # Create SOAP envelope for checkDeployStatus
        soap_envelope = f"""<?xml version="1.0" encoding="UTF-8"?>
<soapenv:Envelope xmlns:soapenv="{soapenv_ns}" xmlns:met="{metadata_ns}">
   <soapenv:Header>
      <met:SessionHeader>
         <met:sessionId>{session_id}</met:sessionId>
      </met:SessionHeader>
   </soapenv:Header>
   <soapenv:Body>
      <met:checkDeployStatus>
         <met:asyncProcessId>{deploy_id}</met:asyncProcessId>
         <met:includeDetails>true</met:includeDetails>
      </met:checkDeployStatus>
   </soapenv:Body>
</soapenv:Envelope>
"""
        headers = {"Content-Type": "text/xml", "SOAPAction": "checkDeployStatus"}

        max_polls = 20
        poll_interval = 5  # Increased polling interval

        for i in range(max_polls):
            logger.info(f"Polling deployment status for {deploy_id} (Attempt {i + 1}/{max_polls})")
            response = requests.post(soap_url, headers=headers, data=soap_envelope.encode('utf-8')) # Ensure UTF-8 encoding

            if response.status_code == 200:
                try:
                    root = ET.fromstring(response.content) # Use response.content for binary data
                    result_node = root.find('.//met:result', ns_map)
                    if result_node is None:
                        logger.error(f"Could not find result node in deploy status response: {response.text}")
                        return {"success": False, "error": "Invalid SOAP response format", "details": response.text[:500]}

                    done_elem = result_node.find('met:done', ns_map)
                    status_elem = result_node.find('met:status', ns_map)
                    status = status_elem.text if status_elem is not None else "Unknown"
                    logger.info(f"Deployment status: {status}")

                    if done_elem is not None and done_elem.text.lower() == "true":
                        success_elem = result_node.find('met:success', ns_map)
                        if success_elem is not None and success_elem.text.lower() == "true":
                            logger.info(f"Deployment {deploy_id} completed successfully.")
                            return {"success": True, "message": "Deployment completed successfully", "status": status}
                        else:
                            logger.error(f"Deployment {deploy_id} failed. Status: {status}")
                            error_messages = []
                            component_failures = result_node.findall('.//met:details/met:componentFailures', ns_map)
                            for failure in component_failures:
                                problem = failure.find('met:problem', ns_map)
                                problem_type = failure.find('met:problemType', ns_map)
                                component_type = failure.find('met:componentType', ns_map)
                                component_name = failure.find('met:fullName', ns_map)
                                error_text = (f"{component_type.text if component_type is not None else ''} "
                                              f"{component_name.text if component_name is not None else ''}: "
                                              f"{problem_type.text if problem_type is not None else 'Error'} - "
                                              f"{problem.text if problem is not None else 'Unknown failure'}")
                                error_messages.append(error_text)
                                logger.error(f"Deployment Failure Detail: {error_text}")

                            if not error_messages:
                                status_msg_elem = result_node.find('met:statusMessage', ns_map)
                                error_messages.append(f"Deployment failed: {status_msg_elem.text if status_msg_elem is not None else status}")

                            return {"success": False, "error": "Deployment failed", "status": status, "details": error_messages}
                    else:
                        # Not done yet, wait
                        time.sleep(poll_interval)
                except ET.ParseError as xml_err:
                     logger.error(f"Error parsing deployment status XML response: {xml_err}\nResponse text: {response.text[:500]}")
                     return {"success": False, "error": "Failed to parse deployment status response", "details": str(xml_err)}
            else:
                logger.error(f"Error checking deployment status. HTTP Status: {response.status_code}, Response: {response.text[:500]}")
                return {"success": False, "error": f"HTTP Error checking deployment status: {response.status_code}", "details": response.text[:500]}

        # Timeout
        logger.warning(f"Deployment {deploy_id} status check timed out after {max_polls * poll_interval} seconds.")
        return {"success": False, "error": "Deployment status check timed out"}

    except requests.exceptions.RequestException as req_err:
        logger.error(f"Network error checking deployment status: {str(req_err)}")
        return {"success": False, "error": "Network error during deployment status check", "details": str(req_err)}
    except Exception as e:
        logger.exception(f"Unexpected error checking deployment status for {deploy_id}: {str(e)}")
        return {"success": False, "error": "Unexpected error during deployment status check", "details": str(e)}

def batch_deploy_metadata(components: List[Dict[str, str]]) -> Dict[str, Any]:
    """
    Deploys multiple metadata components using the Metadata API (SOAP).
    (Corrected: Removed the <runTests> tag from the SOAP envelope when testLevel is NoTestRun)

    Args:
        components: A list of dictionaries, each representing a metadata component.
                    Required keys: 'type' (e.g., 'CustomObject'), 'fullName' (API name), 'content' (XML string).

    Returns:
        A dictionary containing the deployment result ('success': bool, 'deploymentId', 'message'/'error', 'details').
    """
    if not check_connection():
        return {"success": False, "error": "Not connected to Salesforce. Check environment variables or logs."} # Updated error message

    instance_url = get_sf_instance_url()
    session_id = get_sf_session_id()
    if not instance_url or not session_id:
         # This should theoretically not happen if check_connection succeeded, but good practice
         logger.error("Connection check passed but failed to get instance/session ID.")
         return {"success": False, "error": "Salesforce connection details missing unexpectedly."}

    # Directory mappings for common metadata types
    dir_map = {
        "CustomObject": "objects",
        "CustomField": "objects", # Fields are nested under objects
        "ApexClass": "classes",
        "ApexTrigger": "triggers",
        "Layout": "layouts",
        # Add other types as needed
    }

    try:
        with tempfile.TemporaryDirectory() as tmpdir:
            package_manifest = ET.Element("Package", xmlns="http://soap.sforce.com/2006/04/metadata")
            types_map = {}

            # Prepare component files and package.xml structure
            for component in components:
                comp_type = component.get("type")
                comp_fullname = component.get("fullName")
                comp_content = component.get("content")

                if not all([comp_type, comp_fullname, comp_content]):
                    logger.error(f"Skipping invalid component data: {component}")
                    continue

                if comp_type not in types_map:
                    types_map[comp_type] = ET.SubElement(package_manifest, "types")
                    ET.SubElement(types_map[comp_type], "name").text = comp_type

                # Member for field is Object API name, others use their full name
                member_name = comp_fullname.split('.')[0] if comp_type == "CustomField" else comp_fullname
                # Check if member already exists for this type to avoid duplicates
                existing_members = {elem.text for elem in types_map[comp_type].findall("members")}
                if member_name not in existing_members:
                    ET.SubElement(types_map[comp_type], "members").text = member_name

                # --- File Path Logic ---
                target_dir = dir_map.get(comp_type)
                if not target_dir:
                     logger.error(f"Directory mapping not found for type: {comp_type}. Skipping {comp_fullname}.")
                     continue

                if comp_type == "CustomObject":
                    obj_dir_path = os.path.join(tmpdir, target_dir)
                    os.makedirs(obj_dir_path, exist_ok=True)
                    file_path = os.path.join(obj_dir_path, f"{comp_fullname}.object")
                    zip_arcname = f"{target_dir}/{comp_fullname}.object"
                elif comp_type == "CustomField":
                    object_name, field_name = comp_fullname.split('.')
                    field_dir_path = os.path.join(tmpdir, target_dir, object_name, "fields")
                    os.makedirs(field_dir_path, exist_ok=True)
                    file_path = os.path.join(field_dir_path, f"{field_name}.field-meta.xml")
                    zip_arcname = f"{target_dir}/{object_name}/fields/{field_name}.field-meta.xml"
                elif comp_type in ["ApexClass", "ApexTrigger"]:
                    ext = ".cls" if comp_type == "ApexClass" else ".trigger"
                    meta_ext = ".cls-meta.xml" if comp_type == "ApexClass" else ".trigger-meta.xml"
                    code_dir_path = os.path.join(tmpdir, target_dir)
                    os.makedirs(code_dir_path, exist_ok=True)
                    file_path = os.path.join(code_dir_path, f"{comp_fullname}{ext}")
                    zip_arcname = f"{target_dir}/{comp_fullname}{ext}"
                    meta_file_path = os.path.join(code_dir_path, f"{comp_fullname}{meta_ext}")
                    meta_zip_arcname = f"{target_dir}/{comp_fullname}{meta_ext}"
                    with open(meta_file_path, "w", encoding="utf-8") as f_meta:
                        meta_content = component.get('meta_content', f"""<?xml version="1.0" encoding="UTF-8"?>
<{comp_type} xmlns="http://soap.sforce.com/2006/04/metadata">
    <apiVersion>{SALESFORCE_API_VERSION}</apiVersion>
    <status>Active</status>
</{comp_type}>""")
                        f_meta.write(meta_content)
                    component['meta_file_path'] = meta_file_path
                    component['meta_zip_arcname'] = meta_zip_arcname
                else:
                     logger.warning(f"ZIP structure not fully implemented for type: {comp_type}. Using basic structure.")
                     file_dir_path = os.path.join(tmpdir, target_dir)
                     os.makedirs(file_dir_path, exist_ok=True)
                     extension = f".{comp_type.lower()}"
                     file_path = os.path.join(file_dir_path, f"{comp_fullname}{extension}")
                     zip_arcname = f"{target_dir}/{comp_fullname}{extension}"

                # --- Write Content ---
                try:
                    with open(file_path, "w", encoding="utf-8") as f:
                        f.write(comp_content)
                    logger.info(f"Written component file to {file_path} for ZIP path {zip_arcname}")
                except IOError as io_err:
                    logger.error(f"Failed to write component file {file_path}: {io_err}")
                    continue

                # Store paths for zipping
                component['file_path'] = file_path
                component['zip_arcname'] = zip_arcname

            # Add version to package.xml
            ET.SubElement(package_manifest, "version").text = SALESFORCE_API_VERSION

            # Write package.xml
            package_xml_str = prettify_xml(package_manifest)
            package_path = os.path.join(tmpdir, "package.xml")
            with open(package_path, "w", encoding="utf-8") as f:
                f.write(package_xml_str)
            logger.info(f"Generated package.xml:\n{package_xml_str}")

            # Create ZIP file
            zip_path = os.path.join(tmpdir, "metadata.zip")
            with zipfile.ZipFile(zip_path, "w", zipfile.ZIP_DEFLATED) as zipf:
                zipf.write(package_path, arcname="package.xml")
                for component in components:
                    if 'file_path' in component and 'zip_arcname' in component:
                        if os.path.exists(component['file_path']):
                            zipf.write(component['file_path'], arcname=component['zip_arcname'])
                        else:
                             logger.warning(f"Component file {component['file_path']} not found for zipping.")
                    if 'meta_file_path' in component and 'meta_zip_arcname' in component:
                         if os.path.exists(component['meta_file_path']):
                             zipf.write(component['meta_file_path'], arcname=component['meta_zip_arcname'])
                         else:
                            logger.warning(f"Component meta file {component['meta_file_path']} not found for zipping.")

            logger.info(f"Created deployment package at {zip_path}")
            logger.info(f"ZIP contents: {zipf.namelist()}")

            # Read ZIP file content for SOAP request
            with open(zip_path, "rb") as f:
                zip_content_bytes = f.read()
            zip_content = base64.b64encode(zip_content_bytes).decode("utf-8")

            # --- SOAP Deploy Call ---
            soap_url = f"https://{instance_url}/services/Soap/m/{SALESFORCE_API_VERSION}"
            metadata_ns = "http://soap.sforce.com/2006/04/metadata"
            soapenv_ns = "http://schemas.xmlsoap.org/soap/envelope/"
            ns_map = {"met": metadata_ns, "soapenv": soapenv_ns}

            # *** CORRECTED SOAP ENVELOPE: Removed <met:runTests> when testLevel is NoTestRun ***
            soap_envelope = f"""<soapenv:Envelope xmlns:soapenv="{soapenv_ns}" xmlns:met="{metadata_ns}">
   <soapenv:Header>
      <met:SessionHeader>
         <met:sessionId>{session_id}</met:sessionId>
      </met:SessionHeader>
   </soapenv:Header>
   <soapenv:Body>
      <met:deploy>
         <met:ZipFile>{zip_content}</met:ZipFile>
         <met:DeployOptions>
            <met:allowMissingFiles>false</met:allowMissingFiles>
            <met:autoUpdatePackage>false</met:autoUpdatePackage>
            <met:checkOnly>false</met:checkOnly>
            <met:ignoreWarnings>false</met:ignoreWarnings>
            <met:performRetrieve>false</met:performRetrieve>
            <met:purgeOnDelete>false</met:purgeOnDelete>
            <met:rollbackOnError>true</met:rollbackOnError>
            {""}
            <met:testLevel>NoTestRun</met:testLevel>
            <met:singlePackage>true</met:singlePackage>
         </met:DeployOptions>
      </met:deploy>
   </soapenv:Body>
</soapenv:Envelope>
"""
            headers = {"Content-Type": "text/xml; charset=utf-8", "SOAPAction": "deploy"}

            logger.info(f"Sending batch deploy request for {len(components)} components to {soap_url}")
            response = requests.post(soap_url, headers=headers, data=soap_envelope.encode('utf-8'))

            logger.info(f"Deploy request response status: {response.status_code}")
            if 200 <= response.status_code < 300:
                try:
                    root = ET.fromstring(response.content)
                    deploy_id_elem = root.find('.//met:result/met:id', ns_map)

                    if deploy_id_elem is not None:
                        deploy_id = deploy_id_elem.text
                        logger.info(f"Deployment submitted successfully. Deployment ID: {deploy_id}")
                        # Poll for deployment status
                        status_result = check_deployment_status(instance_url, session_id, deploy_id)
                        status_result["deploymentId"] = deploy_id # Add deploy ID to result
                        return status_result
                    else:
                        logger.error(f"Failed to extract deployment ID from SOAP response: {response.text[:500]}")
                        return {"success": False, "error": "Failed to get deployment ID", "details": response.text[:1000]}
                except ET.ParseError as xml_err:
                     logger.error(f"Error parsing deploy SOAP response: {xml_err}\nResponse text: {response.text[:500]}")
                     return {"success": False, "error": "Failed to parse deploy SOAP response", "details": str(xml_err)}
            else:
                logger.error(f"Failed to submit deployment. HTTP Status: {response.status_code}, Response: {response.text[:500]}")
                return {"success": False, "error": f"Failed to deploy batch (HTTP {response.status_code})", "details": response.text[:1000]}

    except Exception as e:
        logger.exception(f"Unexpected error during batch deployment: {str(e)}")
        return {"success": False, "error": "Unexpected error during batch deployment", "details": str(e)}

# --- MCP Tools ---

# REMOVED: The connect_to_salesforce tool is no longer needed.
# @mcp.tool()
# def connect_to_salesforce(username: str, password: str, security_token: str) -> Dict[str, Any]:
#    ... (Original function code deleted) ...

@mcp.tool()
def get_salesforce_connection_status() -> Dict[str, Any]:
    """
    Checks the status of the current Salesforce connection, attempting to connect automatically if needed.

    Verifies if a connection is active and valid before attempting operations.
    Uses environment variables (SF_USERNAME, SF_PASSWORD, SF_SECURITY_TOKEN) for connection.

    Returns:
        A dictionary indicating connection status ('connected': bool), instance URL, and Org ID.
        If connection fails or is invalid, returns 'connected': False and an error message.
    """
    if not check_connection():
        # check_connection now handles the auto-connect attempt and logging
        return {
            "connected": False,
            "message": "Failed to establish automatic Salesforce connection. Check environment variables (SF_USERNAME, SF_PASSWORD, SF_SECURITY_TOKEN) and server logs.",
            "error": "Connection unavailable."
        }

    # If check_connection() returned True, sf should be valid, but we double-check instance/session
    instance_url = get_sf_instance_url()
    session_id = get_sf_session_id()

    if not instance_url or not session_id:
        # This state should be unlikely if check_connection passed, but handle defensively
        logger.error("Connection check reported success, but instance URL or Session ID is missing.")
        global sf # Access global sf to reset it
        sf = None # Reset the connection state as it's inconsistent
        return {
            "connected": False,
            "message": "Internal error: Connection state inconsistent.",
            "error": "Instance/Session ID missing after connect."
        }

    # Attempt a simple query to verify the *existing* session is still active
    try:
        org_info = sf.query("SELECT Id, Name FROM Organization LIMIT 1")
        org_id_from_session = "Unknown"
        if '!' in session_id:
            org_id_from_session = session_id.split('!')[0] # Extract OrgId from session

        org_name = "Unknown"
        if org_info and org_info.get("records"):
             org_name = org_info["records"][0].get("Name", "N/A")

        logger.info(f"Salesforce connection active for Org: {org_name} ({org_id_from_session}) on instance {instance_url}")
        return {
            "connected": True,
            "message": f"Connected to Salesforce Org: {org_name} ({org_id_from_session})",
            "instance": instance_url,
            "org_id": org_id_from_session,
            "org_name": org_name,
        }
    except Exception as e:
        logger.error(f"Salesforce connection test failed (session might be expired): {str(e)}")
        # Don't reset global sf here automatically, let the *next* tool call trigger a reconnect via check_connection
        return {
            "connected": False,
            "message": f"Connection test failed (session may be expired or invalid): {str(e)}",
            "error": str(e)
        }

@mcp.tool()
def describe_metadata_types() -> List[Dict[str, Any]]:
    """
    Lists available metadata types in the connected Salesforce org using the REST Metadata API.

    Use this tool to discover the valid 'xmlName' for different metadata types (e.g., 'CustomObject',
    'ApexClass', 'Profile') which can then be used as input for the 'list_metadata' tool.
    Falls back to a predefined common list if the API call fails. Requires automatic connection.

    Returns:
        A list of dictionaries, each describing a metadata type (keys like 'xmlName',
        'directoryName', 'description', 'inFolder'), or a list containing an error dictionary.
    """
    if not check_connection():
        return [{"error": "Not connected to Salesforce. Check environment variables or logs."}]

    instance_url = get_sf_instance_url()
    session_id = get_sf_session_id()
    if not instance_url or not session_id:
         return [{"error": "Salesforce connection details missing unexpectedly."}] # Should not happen

    try:
        describe_url = f"https://{instance_url}/services/data/v{SALESFORCE_API_VERSION}/metadata/describe"
        headers = {
            "Authorization": f"Bearer {session_id}",
            "Content-Type": "application/json",
        }
        logger.info(f"Requesting metadata description from: {describe_url}")
        response = requests.get(describe_url, headers=headers, params={'apiVersion': SALESFORCE_API_VERSION})

        if 200 <= response.status_code < 300:
            metadata_info = response.json()
            logger.info(f"Successfully retrieved {len(metadata_info.get('metadataObjects', []))} metadata types.")
            return [
                {
                    "xmlName": meta_type.get("xmlName"),
                    "directoryName": meta_type.get("directoryName"),
                    "inFolder": meta_type.get("inFolder", False),
                    "metaFile": meta_type.get("metaFile", False),
                    "suffix": meta_type.get("suffix"),
                    "description": meta_type.get("description", "No description available"),
                }
                for meta_type in metadata_info.get("metadataObjects", [])
            ]
        else:
            logger.warning(f"Failed to describe metadata types via API (Status: {response.status_code}). Falling back to predefined list. Response: {response.text[:200]}")
            # Fallback to a predefined list
            return [
                {"xmlName": "CustomObject", "directoryName": "objects", "description": "Custom Objects"},
                {"xmlName": "CustomField", "directoryName": "fields", "description": "Custom Fields"},
                {"xmlName": "ApexClass", "directoryName": "classes", "description": "Apex Classes"},
                {"xmlName": "Layout", "directoryName": "layouts", "description": "Page Layouts"},
                {"xmlName": "Profile", "directoryName": "profiles", "description": "User Profiles"},
                {"xmlName": "Report", "directoryName": "reports", "description": "Reports"},
                {"xmlName": "Dashboard", "directoryName": "dashboards", "description": "Dashboards"},
            ]
    except requests.exceptions.RequestException as req_err:
         logger.error(f"Network error describing metadata types: {str(req_err)}")
         return [{"error": f"Network error: {str(req_err)}"}]
    except Exception as e:
        logger.exception("Unexpected error describing metadata types.")
        return [{"error": f"Unexpected error: {str(e)}"}]

@mcp.tool()
def list_metadata(metadata_type: str, folder: Optional[str] = None) -> List[Dict[str, Any]]:
    """
    Lists members (specific components) of a given metadata type using the Metadata API (listMetadata).

    Useful for finding existing components (e.g., all Custom Objects, specific Apex classes)
    before creating new ones or modifying existing ones. Use 'describe_metadata_types'
    to find the correct 'metadata_type' value (xmlName). Requires automatic connection.

    Args:
        metadata_type: The XML name of the metadata type (e.g., 'CustomObject', 'ApexClass', 'Report').
        folder: Optional. The folder name ONLY for folder-based metadata types (e.g., 'Report',
                'Dashboard', 'EmailTemplate'). Do not use for types like 'CustomObject'.

    Returns:
        A list of dictionaries, each describing a metadata member (keys like 'fullName',
        'type', 'lastModifiedDate'), or a list containing an error dictionary. Returns an empty
        list if no members are found.
    """
    if not check_connection():
        return [{"error": "Not connected to Salesforce. Check environment variables or logs."}]

    instance_url = get_sf_instance_url()
    session_id = get_sf_session_id()
    if not instance_url or not session_id:
         return [{"error": "Salesforce connection details missing unexpectedly."}]

    try:
        soap_url = f"https://{instance_url}/services/Soap/m/{SALESFORCE_API_VERSION}"
        metadata_ns = "http://soap.sforce.com/2006/04/metadata"
        soapenv_ns = "http://schemas.xmlsoap.org/soap/envelope/"
        ns_map = {"met": metadata_ns, "soapenv": soapenv_ns}

        queries_xml_list = [f"<met:type>{metadata_type}</met:type>"]
        if folder:
            queries_xml_list.append(f"<met:folder>{folder}</met:folder>")

        soap_envelope = f"""<?xml version="1.0" encoding="UTF-8"?>
<soapenv:Envelope xmlns:soapenv="{soapenv_ns}" xmlns:met="{metadata_ns}">
   <soapenv:Header>
      <met:SessionHeader>
         <met:sessionId>{session_id}</met:sessionId>
      </met:SessionHeader>
   </soapenv:Header>
   <soapenv:Body>
      <met:listMetadata>
         <met:queries>
            {''.join(queries_xml_list)}
         </met:queries>
         <met:asOfVersion>{SALESFORCE_API_VERSION}</met:asOfVersion>
      </met:listMetadata>
   </soapenv:Body>
</soapenv:Envelope>
"""
        headers = {"Content-Type": "text/xml; charset=utf-8", "SOAPAction": "listMetadata"}

        logger.info(f"Requesting listMetadata for type: {metadata_type} (Folder: {folder})")
        response = requests.post(soap_url, headers=headers, data=soap_envelope.encode('utf-8'))

        if 200 <= response.status_code < 300:
            try:
                root = ET.fromstring(response.content)
                results = []
                for member in root.findall('.//met:result', ns_map):
                    member_info = {}
                    for child in member:
                        # Extract tag name without namespace
                        tag_name = child.tag.split('}', 1)[-1] if '}' in child.tag else child.tag
                        member_info[tag_name] = child.text
                    if member_info: # Only add if we found some info
                        results.append(member_info)
                logger.info(f"Found {len(results)} members for type {metadata_type}.")
                return results
            except ET.ParseError as xml_err:
                logger.error(f"Error parsing listMetadata SOAP response: {xml_err}\nResponse: {response.text[:500]}")
                return [{"error": "Failed to parse listMetadata response", "details": str(xml_err)}]
        else:
            logger.error(f"listMetadata request failed. HTTP Status: {response.status_code}, Response: {response.text[:500]}")
            # Try to parse Salesforce faultstring
            fault_detail = response.text[:500]
            try:
                root = ET.fromstring(response.content)
                fault_elem = root.find('.//faultstring')
                if fault_elem is not None:
                    fault_detail = fault_elem.text
            except ET.ParseError:
                pass # Keep original text if parse fails
            return [{"error": f"listMetadata failed (HTTP {response.status_code})", "details": fault_detail}]

    except requests.exceptions.RequestException as req_err:
        logger.error(f"Network error during listMetadata: {str(req_err)}")
        return [{"error": f"Network error: {str(req_err)}"}]
    except Exception as e:
        logger.exception(f"Unexpected error during listMetadata for {metadata_type}.")
        return [{"error": f"Unexpected error: {str(e)}"}]

@mcp.tool()
def describe_object(object_name: str) -> Dict[str, Any]:
    """
    Retrieves detailed information about a specific Salesforce object (Standard or Custom).

    Provides essential details about fields, relationships, record types, etc.
    Requires automatic connection.

    Args:
        object_name: The API name of the object (e.g., 'Account', 'MyObject__c').

    Returns:
        A dictionary containing the detailed object description, or an error dictionary.
    """
    if not check_connection():
        return {"error": "Not connected to Salesforce. Check environment variables or logs."}

    try:
        logger.info(f"Describing object: {object_name}")
        # Use simple-salesforce's dynamic attribute access for describe
        description = sf.__getattr__(object_name).describe()
        logger.info(f"Successfully described object: {object_name}")
        # Return the full description; LLM can parse/summarize if needed
        return description
    except Exception as e:
        logger.exception(f"Error describing object {object_name}: {str(e)}")
        # Add more specific error handling if needed (e.g., object not found)
        error_msg = str(e)
        if "The requested resource does not exist" in error_msg:
             error_msg = f"Object '{object_name}' not found or accessible."
        return {"error": f"Failed to describe object {object_name}: {error_msg}"}


@mcp.tool()
def create_custom_field(
    object_name: str,
    field_label: str,
    field_type: str = "Text",
    required: bool = False,
    unique: bool = False,
    external_id: bool = False,
    description: str = "",
    length: Optional[int] = None,
    precision: Optional[int] = None,
    scale: Optional[int] = None,
    picklist_values: Optional[List[str]] = None,
    reference_to: Optional[str] = None,
    relationship_label: Optional[str] = None,
    default_value: Optional[Any] = None
) -> Dict[str, Any]:
    """
    Creates a *single* custom field on a Salesforce object using the Tooling API.

    Faster for single fields, but use 'create_data_model' for multiple/transactional changes.
    Requires automatic connection.

    Args:
        object_name: API name of the object (e.g., 'Account', 'MyObject__c').
        field_label: Human-readable label.
        field_type: Salesforce type (Text, Number, Date, Checkbox, Picklist, Lookup, etc.). Defaults to "Text".
        required: Is field required? Defaults to False.
        unique: Is field unique? Defaults to False.
        external_id: Is field External ID? Defaults to False.
        description: Field description.
        length: Length for Text/LongTextArea/Html.
        precision: Total digits for Number/Currency/Percent.
        scale: Decimal places for Number/Currency/Percent.
        picklist_values: List of values for Picklist.
        reference_to: Parent object API name for Lookup/MasterDetail.
        relationship_label: Related list label for Lookup/MasterDetail.
        default_value: Default value (check Salesforce syntax).

    Returns:
        Success/failure dictionary with field API name/ID or error details.
    """
    if not check_connection():
        return {"success": False, "error": "Not connected to Salesforce. Check environment variables or logs."}

    instance_url = get_sf_instance_url()
    session_id = get_sf_session_id()
    if not instance_url or not session_id:
         return {"success": False, "error": "Salesforce connection details missing unexpectedly."}

    try:
        field_api_name = to_api_name(field_label, is_object=False)
        logger.info(f"Attempting to create field '{field_label}' ({field_api_name}) on object '{object_name}' via Tooling API")

        # Map common type names to Salesforce Tooling API types
        field_type_map = {
            "text": "Text", "string": "Text",
            "number": "Number", "integer": "Number", "double": "Number",
            "currency": "Currency",
            "date": "Date",
            "datetime": "DateTime", "timestamp": "DateTime",
            "checkbox": "Checkbox", "boolean": "Checkbox",
            "picklist": "Picklist",
            "multipicklist": "MultiselectPicklist",
            "lookup": "Lookup",
            "masterdetail": "MasterDetail",
            "textarea": "TextArea", # Maps to Text Area(255) in UI
            "longtextarea": "LongTextArea",
            "richtextarea": "Html", "html": "Html",
            "email": "Email",
            "phone": "Phone",
            "url": "Url",
            "percent": "Percent",
        }
        sf_field_type = field_type_map.get(field_type.lower(), "Text")

        metadata: Dict[str, Any] = {
            "label": field_label,
            "type": sf_field_type,
            "required": required,
            "unique": unique,
            "externalId": external_id,
        }
        if description:
            metadata["description"] = description
            metadata["inlineHelpText"] = description

        # Type-specific properties
        if sf_field_type == "Text": metadata["length"] = min(max(1, length if length is not None else 255), 255)
        elif sf_field_type in ["LongTextArea", "Html"]: metadata["length"] = min(max(256, length if length is not None else 32768), 131072); metadata["visibleLines"] = 3
        elif sf_field_type in ["Number", "Currency", "Percent"]:
             eff_prec = min(max(1, precision if precision is not None else 18), 18)
             eff_scale = min(max(0, scale if scale is not None else 0), eff_prec)
             metadata["precision"] = eff_prec; metadata["scale"] = eff_scale
        elif sf_field_type == "Checkbox": metadata["defaultValue"] = bool(default_value) if default_value is not None else False
        elif sf_field_type == "Picklist" and picklist_values:
             metadata["valueSet"] = {"restricted": True, "valueSetDefinition": {"sorted": False, "value": [{"fullName": re.sub(r"[^a-zA-Z0-9_]", "_", val).strip("_")[:40], "label": val, "default": False} for val in picklist_values]}}
        elif sf_field_type in ["Lookup", "MasterDetail"] and reference_to:
             metadata["referenceTo"] = reference_to
             rel_label = relationship_label if relationship_label else f"{field_label}s"
             rel_name = re.sub(r'[^a-zA-Z0-9]', '', rel_label) + ("_R" if sf_field_type == "Lookup" else "_MDR") # Basic unique suffix idea
             metadata["relationshipLabel"] = rel_label
             metadata["relationshipName"] = rel_name[:40] # Ensure rel name is within limits
             if sf_field_type == "MasterDetail": metadata["writeRequiresMasterRead"] = False

        if default_value is not None and sf_field_type not in ["Checkbox"]: metadata["defaultValue"] = str(default_value)

        tooling_api_url = f"https://{instance_url}/services/data/v{SALESFORCE_API_VERSION}/tooling/sobjects/CustomField"
        headers = {"Authorization": f"Bearer {session_id}", "Content-Type": "application/json"}
        payload = {"FullName": f"{object_name}.{field_api_name}", "Metadata": metadata}

        logger.info(f"Sending Tooling API Request to {tooling_api_url}")
        logger.debug(f"Request Payload: {json.dumps(payload, indent=2)}")
        response = requests.post(tooling_api_url, headers=headers, json=payload)

        logger.info(f"Tooling API Response Status: {response.status_code}")
        if 200 <= response.status_code < 300:
            result = response.json()
            logger.info(f"Successfully created field {field_api_name} via Tooling API. Result: {result}")
            return {"success": True, "message": f"Created field '{field_label}' ({field_api_name}) on {object_name}.", "apiName": field_api_name, "id": result.get("id")}
        else:
            error_details = f"Status Code: {response.status_code}"
            try: error_json = response.json(); error_details = json.dumps(error_json, indent=2)
            except json.JSONDecodeError: error_details = response.text
            logger.error(f"Tooling API Error Response: {error_details[:500]}")
            return {"success": False, "error": f"Failed to create field '{field_label}' via Tooling API.", "details": error_details}

    except requests.exceptions.RequestException as req_err: logger.error(f"Network error (Tooling): {str(req_err)}"); return {"success": False, "error": f"Network error: {str(req_err)}"}
    except Exception as e: logger.exception(f"Unexpected error creating field {field_label}: {str(e)}"); return {"success": False, "error": f"Unexpected error: {str(e)}"}

@mcp.tool()
def create_custom_object(
    object_label: str,
    plural_label: Optional[str] = None,
    description: str = "",
    enable_activities: bool = True,
    enable_reports: bool = True,
    sharing_model: str = "ReadWrite",
    deployment_status: str = "Deployed",
    name_field_label: str = "Name",
    name_field_type: str = "Text",
) -> Dict[str, Any]:
    """
    Creates a *single* new custom object using the Metadata API (via batch deploy helper).

    Preferred for object creation. Requires automatic connection.

    Args:
        object_label: Singular label (e.g., "Invoice").
        plural_label: Optional plural label (e.g., "Invoices"). Defaults to label + 's'.
        description: Optional description.
        enable_activities: Allow Tasks/Events? Defaults to True.
        enable_reports: Allow reporting? Defaults to True.
        sharing_model: Org-wide sharing (ReadWrite, Read, Private, etc.). Defaults to "ReadWrite".
        deployment_status: Deployed or InDevelopment. Defaults to "Deployed".
        name_field_label: Label for Name field. Defaults to "Name".
        name_field_type: Type for Name field ('Text' or 'AutoNumber'). Defaults to "Text".

    Returns:
        Deployment result summary (success, deploymentId, status, apiName).
    """
    if not check_connection():
        return {"success": False, "error": "Not connected to Salesforce. Check environment variables or logs."} # Consistent error
    try:
        api_name = to_api_name(object_label, is_object=True); eff_plural_label = plural_label if plural_label else object_label + "s"
        logger.info(f"Preparing object '{object_label}' ({api_name}) via Metadata API.")
        object_root = ET.Element("CustomObject", xmlns="http://soap.sforce.com/2006/04/metadata")
        ET.SubElement(object_root, "label").text = object_label; ET.SubElement(object_root, "pluralLabel").text = eff_plural_label
        if description: ET.SubElement(object_root, "description").text = description
        name_field = ET.SubElement(object_root, "nameField"); ET.SubElement(name_field, "label").text = name_field_label
        if name_field_type == "AutoNumber": ET.SubElement(name_field, "type").text = "AutoNumber"; ET.SubElement(name_field, "displayFormat").text = f"{object_label[:3].upper()}-{{0000}}"; ET.SubElement(name_field, "startingNumber").text = "1"; ET.SubElement(name_field, "trackHistory").text = "false"
        else: ET.SubElement(name_field, "type").text = "Text"; ET.SubElement(name_field, "trackHistory").text = "false" # Default Name field history to false
        ET.SubElement(object_root, "deploymentStatus").text = deployment_status; ET.SubElement(object_root, "sharingModel").text = sharing_model
        ET.SubElement(object_root, "enableActivities").text = str(enable_activities).lower(); ET.SubElement(object_root, "enableReports").text = str(enable_reports).lower()
        ET.SubElement(object_root, "enableHistory").text = "false"; ET.SubElement(object_root, "enableBulkApi").text = "true"; ET.SubElement(object_root, "enableStreamingApi").text = "true"; ET.SubElement(object_root, "enableSearch").text = "true"
        object_xml_content = prettify_xml(object_root)
        logger.debug(f"Object XML for {api_name}:\n{object_xml_content}")
        component = {"type": "CustomObject", "fullName": api_name, "content": object_xml_content}
        logger.info(f"Submitting deployment for object {api_name}.")
        deploy_result = batch_deploy_metadata([component]) # Uses the helper that checks connection internally
        if deploy_result.get("success"): logger.info(f"Deployed object {api_name}."); return {"success": True, "message": f"Initiated creation of object '{object_label}' ({api_name}).", "apiName": api_name, "deploymentId": deploy_result.get("deploymentId"), "deploymentStatus": deploy_result.get("status", "Unknown")}
        else: logger.error(f"Failed deploy object {api_name}: {deploy_result.get('error')}"); return {"success": False, "error": f"Failed create object '{object_label}'.", "apiName": api_name, "deploymentId": deploy_result.get("deploymentId"), "details": deploy_result.get("details", "No details")}
    except Exception as e: logger.exception(f"Unexpected error creating object {object_label}: {str(e)}"); return {"success": False, "error": f"Unexpected error: {str(e)}"}


@mcp.tool()
def create_data_model(model_spec: Dict[str, List[Dict[str, Any]]]) -> Dict[str, Any]:
    """
    Creates or updates a *complete data model* (multiple objects, fields, relationships)
    using a single Metadata API deployment. Preferred for defining interconnected components.
    Requires automatic connection.

    *** IMPORTANT: Before calling this tool, you MUST present the generated 'model_spec'
    JSON structure clearly to the user and ask for their explicit confirmation to proceed.
    Only call this tool after receiving user approval. ***

    Args:
        model_spec: Dictionary defining the model structure (objects with fields, relationships). Uses LABELS as input.

    Returns:
        Batch deployment result summary (success, deploymentId, status, messages/errors).
    """
    if not check_connection():
        return {"success": False, "error": "Not connected to Salesforce. Check environment variables or logs."} # Consistent error
    logger.info("Starting complex data model creation via Metadata API.")
    components_to_deploy = []; object_api_names_map: Dict[str, str] = {}
    try:
        logger.info("Processing object definitions...")
        for obj_spec in model_spec.get("objects", []):
            object_label = obj_spec.get("label")
            if not object_label: logger.error("Skip object spec: missing label."); continue
            api_name = to_api_name(object_label, is_object=True); object_api_names_map[object_label] = api_name
            plural_label = obj_spec.get("plural_label", object_label + "s"); description = obj_spec.get("description", ""); name_field_label = obj_spec.get("name_field_label", "Name"); name_field_type = obj_spec.get("name_field_type", "Text")
            # --- Object XML ---
            object_root = ET.Element("CustomObject", xmlns="http://soap.sforce.com/2006/04/metadata"); ET.SubElement(object_root, "label").text = object_label; ET.SubElement(object_root, "pluralLabel").text = plural_label
            if description: ET.SubElement(object_root, "description").text = description
            name_field = ET.SubElement(object_root, "nameField"); ET.SubElement(name_field, "label").text = name_field_label
            if name_field_type == "AutoNumber": ET.SubElement(name_field, "type").text = "AutoNumber"; ET.SubElement(name_field, "displayFormat").text = obj_spec.get("name_field_format", f"{object_label[:3].upper()}-{{0000}}"); ET.SubElement(name_field, "startingNumber").text = str(obj_spec.get("name_field_starting_number", 1)); ET.SubElement(name_field, "trackHistory").text = "false"
            else: ET.SubElement(name_field, "type").text = "Text"; ET.SubElement(name_field, "trackHistory").text = str(obj_spec.get("track_name_field_history", False)).lower()
            ET.SubElement(object_root, "deploymentStatus").text = obj_spec.get("deployment_status", "Deployed"); ET.SubElement(object_root, "sharingModel").text = obj_spec.get("sharing_model", "ReadWrite"); ET.SubElement(object_root, "enableActivities").text = str(obj_spec.get("enable_activities", True)).lower(); ET.SubElement(object_root, "enableReports").text = str(obj_spec.get("enable_reports", True)).lower(); ET.SubElement(object_root, "enableHistory").text = str(obj_spec.get("enable_history", False)).lower(); ET.SubElement(object_root, "enableBulkApi").text = str(obj_spec.get("enable_bulk_api", True)).lower(); ET.SubElement(object_root, "enableStreamingApi").text = str(obj_spec.get("enable_streaming_api", True)).lower(); ET.SubElement(object_root, "enableSearch").text = str(obj_spec.get("enable_search", True)).lower()
            object_xml_content = prettify_xml(object_root); components_to_deploy.append({"type": "CustomObject", "fullName": api_name, "content": object_xml_content}); logger.info(f"Prepared Object: {api_name}")
            # --- Field Components ---
            logger.info(f"Processing fields for {object_label} ({api_name})")
            for field_spec in obj_spec.get("fields", []):
                field_label = field_spec.get("label"); field_type = field_spec.get("type")
                if not field_label or not field_type: logger.error(f"Skip field: missing label/type in {object_label}."); continue
                if field_type.lower() in ["lookup", "masterdetail"]: logger.warning(f"Skip relationship '{field_label}' in object '{object_label}'. Define in 'relationships'."); continue
                field_api_name = to_api_name(field_label, is_object=False); full_field_name = f"{api_name}.{field_api_name}"
                # --- Field XML ---
                field_root = ET.Element("CustomField", xmlns="http://soap.sforce.com/2006/04/metadata"); ET.SubElement(field_root, "fullName").text = field_api_name; ET.SubElement(field_root, "label").text = field_label
                field_type_map = {"text": "Text", "number": "Number", "currency": "Currency", "date": "Date", "datetime": "DateTime", "checkbox": "Checkbox", "picklist": "Picklist", "multipicklist": "MultiselectPicklist", "longtextarea": "LongTextArea", "richtextarea": "Html", "email": "Email", "phone": "Phone", "url": "Url", "percent": "Percent"}
                sf_field_type = field_type_map.get(field_type.lower(), "Text"); ET.SubElement(field_root, "type").text = sf_field_type
                if "description" in field_spec: ET.SubElement(field_root, "description").text = field_spec["description"]
                ET.SubElement(field_root, "required").text = str(field_spec.get("required", False)).lower(); ET.SubElement(field_root, "externalId").text = str(field_spec.get("external_id", False)).lower(); ET.SubElement(field_root, "trackHistory").text = str(field_spec.get("track_history", False)).lower(); ET.SubElement(field_root, "trackTrending").text = str(field_spec.get("track_trending", False)).lower(); ET.SubElement(field_root, "unique").text = str(field_spec.get("unique", False)).lower()
                if sf_field_type == "Text": ET.SubElement(field_root, "length").text = str(min(max(1, field_spec.get("length", 255)), 255))
                elif sf_field_type in ["LongTextArea", "Html"]: ET.SubElement(field_root, "length").text = str(min(max(256, field_spec.get("length", 32768)), 131072)); ET.SubElement(field_root, "visibleLines").text = str(field_spec.get("visible_lines", 3))
                elif sf_field_type in ["Number", "Currency", "Percent"]: precision = field_spec.get("precision", 18); scale_ = field_spec.get("scale", 0); prec_elem = ET.SubElement(field_root, "precision"); prec_elem.text = str(min(max(1, precision), 18)); ET.SubElement(field_root, "scale").text = str(min(max(0, scale_), int(prec_elem.text)))
                elif sf_field_type == "Checkbox": ET.SubElement(field_root, "defaultValue").text = str(field_spec.get("default_value", False)).lower()
                elif sf_field_type == "Picklist" and field_spec.get("picklist_values"):
                     value_set = ET.SubElement(field_root, "valueSet"); value_set_def = ET.SubElement(value_set, "valueSetDefinition"); ET.SubElement(value_set_def, "sorted").text = str(field_spec.get("picklist_sorted", False)).lower()
                     for val in field_spec["picklist_values"]: value_elem = ET.SubElement(value_set_def, "value"); val_api_name = re.sub(r"[^a-zA-Z0-9_]", "_", val).strip("_")[:40]; ET.SubElement(value_elem, "fullName").text = val_api_name; ET.SubElement(value_elem, "default").text = "false"; ET.SubElement(value_elem, "label").text = val
                     ET.SubElement(value_set, 'restricted').text = str(field_spec.get("picklist_restricted", True)).lower()
                if "default_value" in field_spec and sf_field_type not in ["Checkbox"]: ET.SubElement(field_root, "defaultValue").text = str(field_spec["default_value"])
                field_xml_content = prettify_xml(field_root); components_to_deploy.append({"type": "CustomField", "fullName": full_field_name, "content": field_xml_content}); logger.info(f"Prepared Field: {full_field_name}")
        # --- Relationship Field Components ---
        logger.info("Processing relationship definitions...")
        for rel_spec in model_spec.get("relationships", []):
            from_label = rel_spec.get("from_object_label"); to_label = rel_spec.get("to_object_label"); field_label = rel_spec.get("field_label"); rel_type = rel_spec.get("relationship_type", "Lookup")
            if not all([from_label, to_label, field_label]): logger.error("Skip relationship: missing labels."); continue
            from_api_name = object_api_names_map.get(from_label); to_api_name = object_api_names_map.get(to_label)
            if not from_api_name or not to_api_name: logger.error(f"Cannot find API names for relationship: {from_label} -> {to_label}."); continue
            field_api_name = to_api_name(field_label, is_object=False); full_field_name = f"{to_api_name}.{field_api_name}" # Field lives on the 'to' object in our spec (child)
            # --- Relationship XML ---
            field_root = ET.Element("CustomField", xmlns="http://soap.sforce.com/2006/04/metadata"); ET.SubElement(field_root, "fullName").text = field_api_name; ET.SubElement(field_root, "label").text = field_label; ET.SubElement(field_root, "type").text = rel_type
            if "description" in rel_spec: ET.SubElement(field_root, "description").text = rel_spec["description"]
            ET.SubElement(field_root, "required").text = str(rel_spec.get("required", False)).lower(); ET.SubElement(field_root, "externalId").text = "false"; ET.SubElement(field_root, "trackHistory").text = str(rel_spec.get("track_history", False)).lower(); ET.SubElement(field_root, "trackTrending").text = str(rel_spec.get("track_trending", False)).lower()
            ET.SubElement(field_root, "referenceTo").text = from_api_name
            # Correct relationship label/name generation: label on parent, name derived from child
            default_rel_label = to_api_name.replace("__c", "") + "s"; # Plural of the child object by default
            relationship_label = rel_spec.get("relationship_label", default_rel_label);
            relationship_name = re.sub(r'[^a-zA-Z0-9]', '', field_label).replace("__c","") # Name based on the field name itself
            ET.SubElement(field_root, "relationshipLabel").text = relationship_label; ET.SubElement(field_root, "relationshipName").text = relationship_name; ET.SubElement(field_root, "relationshipOrder").text = str(rel_spec.get("relationship_order", 0))
            if rel_type == "MasterDetail": ET.SubElement(field_root, "writeRequiresMasterRead").text = str(rel_spec.get("write_requires_master_read", False)).lower(); delete_constraint = "Cascade" if rel_spec.get("required", True) else "SetNull"; ET.SubElement(field_root, "deleteConstraint").text = rel_spec.get("delete_constraint", delete_constraint); ET.SubElement(field_root, "reparentableMasterDetail").text = str(rel_spec.get("reparentable", False)).lower()
            elif rel_type == "Lookup": delete_constraint = "SetNull" if not rel_spec.get("required", False) else "Restrict"; ET.SubElement(field_root, "deleteConstraint").text = rel_spec.get("delete_constraint", delete_constraint); ET.SubElement(field_root, "required").text = str(rel_spec.get("required", False)).lower() # Re-set required for lookup explicitly
            field_xml_content = prettify_xml(field_root); components_to_deploy.append({"type": "CustomField", "fullName": full_field_name, "content": field_xml_content}); logger.info(f"Prepared Relationship Field: {full_field_name} ({rel_type} to {from_api_name})")
        # --- Deploy all ---
        if not components_to_deploy: return {"success": False, "error": "No valid components in model spec."}
        logger.info(f"Submitting batch deployment for {len(components_to_deploy)} components.")
        deploy_result = batch_deploy_metadata(components_to_deploy) # Uses helper which checks connection
        # --- Process Result ---
        if deploy_result.get("success"): logger.info("Data model deployment successful."); return {"success": True, "message": "Data model creation/update OK.", "deploymentId": deploy_result.get("deploymentId"), "deploymentStatus": deploy_result.get("status", "Succeeded")}
        else: logger.error(f"Data model deployment failed: {deploy_result.get('error')}"); return {"success": False, "error": "Data model creation/update failed.", "deploymentId": deploy_result.get("deploymentId"), "deploymentStatus": deploy_result.get("status", "Failed"), "details": deploy_result.get("details", "No details")}
    except Exception as e: logger.exception(f"Unexpected error creating data model: {str(e)}"); return {"success": False, "error": f"Unexpected error: {str(e)}"}


@mcp.tool()
def generate_soql_query(
    object_name: str,
    fields: Optional[List[str]] = None,
    where_clause: Optional[str] = None,
    order_by: Optional[str] = None,
    limit: Optional[int] = None
) -> Dict[str, Any]:
    """
    Generates a SOQL query string. Does NOT execute. Requires automatic connection (for auto-fields).

    Use 'describe_object' to find valid field API names. Use caution with 'where_clause' and 'order_by'.

    Args:
        object_name: API name of the object.
        fields: Optional list of field API names. Defaults to ['Id'] or auto-discovered fields.
        where_clause: Optional filter condition (e.g., "Industry = 'Technology'"). Use with caution.
        order_by: Optional sorting condition (e.g., "CreatedDate DESC"). Use with caution.
        limit: Optional max number of records.

    Returns:
        Dictionary with 'query' key on success, or error dictionary.
    """
    # Connection check needed *only* if fields are not provided (for describe fallback)
    if not fields:
        if not check_connection():
             # Allow query generation with explicitly listed fields even if not connected
             logger.warning("Cannot auto-describe fields (not connected). Generating query with 'Id' only.")
             selected_fields = ["Id"]
        else:
             try: logger.info(f"Auto-describing fields for {object_name}."); desc = sf.__getattr__(object_name).describe(); selected_fields = [f['name'] for f in desc.get('fields', []) if f.get('name')] or ["Id"]; logger.info(f"Auto-selected {len(selected_fields)} fields.")
             except Exception as desc_err: logger.warning(f"Failed auto-describe {object_name}: {desc_err}. Defaulting to Id."); selected_fields = ["Id"]
    else:
        selected_fields = fields # Use provided fields

    try:
        safe_object_name = re.sub(r"[^a-zA-Z0-9_.]", "", object_name); safe_fields = [re.sub(r"[^a-zA-Z0-9_.,()]", "", f) for f in selected_fields]
        query = f"SELECT {', '.join(safe_fields)} FROM {safe_object_name}"
        if where_clause: logger.warning("WHERE clause provided. Ensure safety."); query += f" WHERE {where_clause}"
        if order_by: logger.warning("ORDER BY clause provided. Ensure safety."); safe_order_by = re.sub(r"[^a-zA-Z0-9_,\s(DESC)(ASC)(NULLS FIRST)(NULLS LAST)]", "", order_by, flags=re.IGNORECASE); query += f" ORDER BY {safe_order_by}"
        if limit is not None and isinstance(limit, int) and limit > 0: query += f" LIMIT {limit}"
        logger.info(f"Generated SOQL: {query}"); return {"success": True, "query": query}
    except Exception as e: logger.exception(f"Error generating SOQL for {object_name}: {str(e)}"); return {"error": f"Failed generate SOQL: {str(e)}"}


@mcp.tool()
def execute_soql_query(query: str) -> List[Dict[str, Any]]:
    """
    Executes a provided SOQL query string. Requires automatic connection.

    Use with queries generated by 'generate_soql_query' or known-safe queries. Retrieves only the first batch of results.

    Args:
        query: The complete SOQL query string.

    Returns:
        List of record dictionaries (attributes removed), empty list if no records, or list with error dictionary on failure.
    """
    if not check_connection():
        return [{"error": "Not connected to Salesforce. Check environment variables or logs."}]
    if not query or "select " not in query.lower() or " from " not in query.lower(): logger.error(f"Invalid SOQL for execution: {query}"); return [{"error": "Invalid SOQL. Must include SELECT and FROM."}]
    logger.info(f"Executing SOQL: {query[:200]}...")
    try:
        result = sf.query(query); records = result.get("records", []); total_size = result.get("totalSize", 0)
        logger.info(f"Query OK. Returned {len(records)} of {total_size} total.")
        if not result.get("done", True): logger.warning(f"Paginated results. Only first {len(records)} returned.")
        cleaned_records = [{k: v for k, v in record.items() if k != 'attributes'} for record in records]
        return cleaned_records
    except Exception as e: logger.exception(f"Error executing SOQL: {str(e)}"); return [{"error": f"Failed execute SOQL: {str(e)}"}]


@mcp.tool()
def generate_test_data(
    object_name: str,
    record_count: int = 5,
    field_mappings: Optional[Dict[str, Any]] = None,
    data_description: Optional[str] = None # NEW Parameter
) -> Dict[str, Any]:
    """
    Generates and inserts a *limited number* of test data records (max 100 per call) using the Bulk API.
    Requires automatic connection.

    Attempts to populate required fields. Use 'data_description' for theme (e.g., "SaaS startups").
    Use 'field_mappings' for specific static values (essential for required lookups). Max 100 records/call.

    Args:
        object_name: API name of the object (e.g., 'Account', 'MyObject__c').
        record_count: Number of test records (defaults to 5, max 100).
        field_mappings: Optional map of field API names to static values (e.g., {"Status__c": "New"}).
        data_description: Optional text description of data theme (e.g., "Manufacturing in Texas").

    Returns:
        Summary dictionary (success/failure counts, IDs, errors). Check 'success' flag and 'errors'.
    """
    if not check_connection():
        return {"error": "Not connected to Salesforce. Check environment variables or logs."}

    MAX_RECORDS = 100 # Add a safeguard
    safe_record_count = max(1, min(record_count, MAX_RECORDS))
    if record_count > MAX_RECORDS:
        logger.warning(f"Requested {record_count} records, but capping at {MAX_RECORDS}.")

    results = {"successful_inserts": 0, "failed_inserts": 0, "record_ids": [], "errors": []}
    if safe_record_count <= 0:
        return {"error": "record_count must be positive."}

    try:
        logger.info(f"Describing object {object_name} to generate test data...")
        desc = sf.__getattr__(object_name).describe()
        all_fields = {field['name']: field for field in desc.get('fields', [])}
        createable_fields = {name: field for name, field in all_fields.items() if field.get('createable')}
        required_fields = {name: field for name, field in createable_fields.items() if not field.get('nillable') and not field.get('defaultedOnCreate')}

        logger.info(f"Found {len(createable_fields)} createable fields, {len(required_fields)} required fields for {object_name}.")
        if not createable_fields:
            return {"error": f"No createable fields found for object {object_name}."}

        # --- Pre-flight Check for field_mappings ---
        if field_mappings:
            for map_field_name in field_mappings.keys():
                if map_field_name not in createable_fields:
                    error_msg = f"Field '{map_field_name}' provided in field_mappings does not exist or is not createable on object '{object_name}'."
                    logger.error(error_msg)
                    return {"error": error_msg, "details": f"Available createable fields: {list(createable_fields.keys())}"}
        # --- End Pre-flight Check ---

        records_to_insert = []
        for i in range(safe_record_count):
            record_data = {}
            record_num = i + 1

            # Apply fixed field mappings first
            if field_mappings:
                record_data.update(field_mappings)

            # --- Simple Themed Data Generation ---
            base_name = f"Test {object_name.replace('__c','')} {record_num}"
            industry = "Technology" # Default
            city = "San Francisco"
            state = "CA"

            if data_description:
                # Very basic theme extraction - enhance as needed
                if "saas" in data_description.lower():
                    industry = "Technology"
                    base_name = random.choice(["Innovate", "Cloudify", "Synergy", "Quantum", "Apex", "Zenith"]) + f" Solutions {record_num}"
                if "bay area" in data_description.lower() or "san francisco" in data_description.lower():
                    city = random.choice(["San Francisco", "Palo Alto", "Mountain View", "San Mateo", "Oakland"])
                    state = "CA"
                elif "manufacturing" in data_description.lower():
                     industry = "Manufacturing"
                     base_name = random.choice(["Acme", "Global", "Precision", "United", "Dynamic"]) + f" Manufacturing {record_num}"
                     city = "Detroit" # Example change
                     state = "MI" # Example change
                # Add more theme keywords...
            # --- End Themed Data Generation ---

            # --- Generate data for required fields if not already mapped ---
            for field_name, field in required_fields.items():
                if field_name in record_data:
                    continue # Skip if already mapped

                field_type = field['type']
                generated_value = None
                field_label_text = field.get('label', field_name)

                # *** START OF REWRITTEN/SIMPLIFIED SECTION ***
                if field_type in ['string', 'textarea', 'url']:
                    if field_name == 'Name':
                        generated_value = base_name[:field.get('length', 80)]
                    elif field_label_text == 'Website':
                        generated_value = f"www.{base_name.split(' ')[0].lower()}{record_num}.test"
                    else:
                        generated_value = f"Test Value {record_num} for {field_label_text}"[:field.get('length', 255)]

                elif field_type == 'email':
                     generated_value = f"info{record_num}@{base_name.split(' ')[0].lower()}.test"

                elif field_type == 'phone':
                    generated_value = f"555-0{record_num:02d}-0000"

                elif field_name == 'BillingCity':
                    if 'BillingCity' not in record_data: generated_value = city
                elif field_name == 'BillingState':
                    if 'BillingState' not in record_data: generated_value = state
                elif field_name == 'ShippingCity': # Added Shipping address example
                     if 'ShippingCity' not in record_data: generated_value = city
                elif field_name == 'ShippingState':
                     if 'ShippingState' not in record_data: generated_value = state
                elif field_name == 'Industry':
                    if 'Industry' not in record_data: generated_value = industry

                elif field_type in ['double', 'int', 'currency', 'percent']:
                    base_num = float(record_num * random.uniform(50, 150))
                    if field_type == 'int':
                        generated_value = int(base_num)
                    elif field_type == 'percent':
                        generated_value = min(base_num / 10.0, 100.0) # Example percent scaling
                        # Apply precision/scale for percent if needed
                        prec = field.get('precision')
                        scale_ = field.get('scale')
                        if prec is not None and scale_ is not None:
                            generated_value = round(generated_value, scale_)
                    else: # double, currency
                        generated_value = base_num
                        # Apply precision/scale
                        prec = field.get('precision')
                        scale_ = field.get('scale')
                        if prec is not None and scale_ is not None:
                            generated_value = round(generated_value, scale_)

                elif field_type == 'date':
                    generated_value = (datetime.date.today() + datetime.timedelta(days=i)).isoformat()

                elif field_type == 'datetime':
                    # Use timezone aware datetime
                    generated_value = (datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(minutes=i)).isoformat()

                elif field_type == 'boolean':
                    generated_value = (i % 2 == 0)

                elif field_type == 'picklist' and field.get('picklistValues'):
                    active_values = [pv['value'] for pv in field['picklistValues'] if pv.get('active')]
                    # Try to use themed value if possible (e.g., Industry)
                    if field_name == 'Industry' and industry in active_values and 'Industry' not in record_data:
                        generated_value = industry
                    elif active_values:
                        generated_value = active_values[i % len(active_values)]
                # *** END OF REWRITTEN/SIMPLIFIED SECTION ***

                # --- Assign generated value or log error ---
                if generated_value is not None:
                    record_data[field_name] = generated_value
                elif field_name not in record_data: # Only error if we couldn't generate AND it wasn't mapped
                    # Could not generate value for a required field
                    error_msg = f"Cannot automatically generate value for required field '{field_name}' (Type: {field_type}). Provide it in field_mappings or make it non-required."
                    logger.error(error_msg)
                    results['errors'].append({"record_index": i, "field_error": error_msg})
                    # Decide whether to stop or continue. Continuing is risky if field is truly required.
                    # For now, we log and continue, hoping SF defaults might apply (unlikely for required)
                    # To be safer, you could uncomment the line below to stop immediately:
                    # return {"error": error_msg, "details": results}

            # Add optional Name field generation if not mapped or required above
            name_field = all_fields.get('Name')
            if 'Name' not in record_data and name_field and name_field.get('createable'):
                 record_data['Name'] = base_name[:name_field.get('length', 80)]

            # Add other common optional fields if not mapped
            if 'Industry' not in record_data and 'Industry' in createable_fields:
                 record_data['Industry'] = industry
            if 'BillingCity' not in record_data and 'BillingCity' in createable_fields:
                 record_data['BillingCity'] = city
            if 'BillingState' not in record_data and 'BillingState' in createable_fields:
                 record_data['BillingState'] = state


            records_to_insert.append(record_data)
            logger.debug(f"Prepared record {i}: {record_data}")


        # --- Bulk Insert ---
        if records_to_insert:
            logger.info(f"Attempting to bulk insert {len(records_to_insert)} records for {object_name}.")
            try:
                # Use bulk API for potentially large inserts
                bulk_results = sf.bulk.__getattr__(object_name).insert(records_to_insert, batch_size=200, use_serial=False) # Smaller batch size might help debug

                success_count = 0
                fail_count = 0
                inserted_ids = []
                bulk_errors = []

                for i, result in enumerate(bulk_results):
                    if isinstance(result, dict) and result.get('success'):
                        success_count += 1
                        inserted_ids.append(result.get('id'))
                    else:
                        fail_count += 1
                        error_detail = {
                            "record_index": i, # Index within this batch attempt
                            "errors": result.get('errors', [{'message': 'Unknown bulk error'}]) if isinstance(result, dict) else [{'message': str(result)}]
                        }
                        bulk_errors.append(error_detail)
                        logger.warning(f"Failed to insert record index {i}: {error_detail['errors']}")

                results['successful_inserts'] = success_count
                results['failed_inserts'] = fail_count
                results['record_ids'] = inserted_ids
                results['errors'].extend(bulk_errors[:10]) # Limit errors in response


            except Exception as bulk_err:
                # Catching broader errors from simple_salesforce bulk
                logger.exception("Bulk insert operation failed.")
                results['errors'].append({"bulk_error": str(bulk_err), "details": "Check logs for specifics. Did a required field fail generation?"})
                # If bulk fails catastrophically, report 0 success/all fail
                results['failed_inserts'] = safe_record_count
                results['successful_inserts'] = 0


        logger.info(f"Test data generation complete for {object_name}. Success: {results['successful_inserts']}, Failed: {results['failed_inserts']}")
        return {
            "success": results['failed_inserts'] == 0 and results['successful_inserts'] > 0,
            "object": object_name,
            "records_attempted": safe_record_count,
            "successful_inserts": results['successful_inserts'],
            "failed_inserts": results['failed_inserts'],
            "record_ids": results['record_ids'],
            "errors": results['errors'] # Return all captured errors
        }

    except Exception as e:
        logger.exception(f"Error generating test data for {object_name}: {str(e)}")
        # Include any partial results gathered before the exception
        return {"error": f"Failed to generate test data: {str(e)}", "details": results}

def field_label(field_dict: Dict[str, Any]) -> str:
    """Helper to get field label or name."""
    return field_dict.get('label', field_dict.get('name', 'Unknown Field'))

@mcp.tool()
def get_org_limits() -> Dict[str, Any]:
    """
    Retrieves current Salesforce organization limits (API usage, storage, etc.) using the REST API.
    Requires automatic connection.

    Returns:
        Dictionary with key limits (DailyApiRequests, DataStorageMB, etc.) or error dictionary.
    """
    if not check_connection():
        return {"error": "Not connected to Salesforce. Check environment variables or logs."}
    instance_url = get_sf_instance_url(); session_id = get_sf_session_id()
    if not instance_url or not session_id: return {"error": "Connection details missing unexpectedly."}
    try:
        limits_url = f"https://{instance_url}/services/data/v{SALESFORCE_API_VERSION}/limits/"; headers = {"Authorization": f"Bearer {session_id}", "Content-Type": "application/json"}
        logger.info("Requesting org limits..."); response = requests.get(limits_url, headers=headers)
        if 200 <= response.status_code < 300:
            limits_data = response.json(); logger.info("Retrieved org limits.")
            simple_limits = {"DailyApiRequests": limits_data.get("DailyApiRequests"), "DataStorageMB": limits_data.get("DataStorageMB"), "FileStorageMB": limits_data.get("FileStorageMB"), "MonthlyPlatformEvents": limits_data.get("MonthlyPlatformEvents")}
            return {"success": True, "limits": simple_limits} # Return simplified version by default
        else: logger.error(f"Failed get limits. Status: {response.status_code}: {response.text[:200]}"); return {"error": f"Failed get limits (HTTP {response.status_code})", "details": response.text[:500]}
    except requests.exceptions.RequestException as req_err: logger.error(f"Network error getting limits: {str(req_err)}"); return {"error": f"Network error: {str(req_err)}"}
    except Exception as e: logger.exception("Unexpected error getting limits."); return {"error": f"Unexpected error: {str(e)}"}


@mcp.tool()
def search_salesforce_documentation(query: str, num_results: int = 5) -> Dict[str, Any]:
    """
    Searches trusted Salesforce documentation sites (architect.salesforce.com, help.salesforce.com)
    using Google Custom Search. Requires GOOGLE_API_KEY and GOOGLE_CSE_ID environment variables.

    Use to find official guidance on features, best practices, or errors.

    Args:
        query: Search query (e.g., "Apex transaction control").
        num_results: Max results (default: 5, max: 10).

    Returns:
        Dictionary with list of results ('title', 'link', 'snippet') or error dictionary.
    """
    if not GOOGLE_LIBS_AVAILABLE: return {"error": "Google API libraries not installed. Cannot perform search."}
    api_key = os.getenv("GOOGLE_API_KEY"); cse_id = os.getenv("GOOGLE_CSE_ID")
    if not api_key or not cse_id: logger.error("Google API Key/CSE ID missing."); return {"error": "Search config missing (API Key/CSE ID)."}
    num = max(1, min(num_results, 10))
    try:
        logger.info(f"Performing Google CSE search: '{query}' (num={num})"); service = build("customsearch", "v1", developerKey=api_key)
        result = service.cse().list(q=query, cx=cse_id, num=num).execute()
        search_items = result.get('items', []); formatted_results = []
        if search_items:
            for item in search_items: formatted_results.append({"title": item.get('title'), "link": item.get('link'), "snippet": item.get('snippet')})
            logger.info(f"Found {len(formatted_results)} results."); return {"success": True, "results": formatted_results}
        else: logger.info(f"No results found."); return {"success": True, "results": [], "message": "No results found."}
    except HttpError as e:
        error_details = e.resp.reason
        try: error_content = json.loads(e.content); error_details = error_content.get('error', {}).get('message', e.resp.reason)
        except Exception: pass
        logger.error(f"Google CSE API error: {e.resp.status} {error_details}"); return {"error": f"Search API error: {error_details}", "status_code": e.resp.status}
    except Exception as e: logger.exception(f"Unexpected search error: {str(e)}"); return {"error": f"Unexpected search error: {str(e)}"}


# --- Server Execution ---

if __name__ == "__main__":
    logger.info("Starting Salesforce MCP Server...")
    logger.info("Attempting initial Salesforce connection on startup via check_connection()...")
    # Optionally attempt connection on startup to fail fast if env vars are wrong
    if not check_connection():
         logger.error("**************************************************************")
         logger.error("*** Initial Salesforce connection FAILED.                  ***")
         logger.error("*** Check SF_USERNAME, SF_PASSWORD, SF_SECURITY_TOKEN env vars ***")
         logger.error("*** and ensure network connectivity to Salesforce.           ***")
         logger.error("**************************************************************")
         # Decide if you want the server to exit or continue without connection
         # exit(1) # Uncomment to exit if initial connection fails
    else:
        logger.info("Initial Salesforce connection successful (or already established).")

    mcp.run()
    logger.info("Salesforce MCP Server stopped.")

# --- END OF MODIFIED FILE ---