# MCP Server for Salesforce Metadata Integration

This Python server implements the Model Context Protocol (MCP) to enable Claude AI to interact with Salesforce metadata through natural language. It provides a comprehensive set of tools for connecting to Salesforce, managing metadata, executing SOQL queries, and generating test data.

## Technical Overview

### Model Context Protocol Architecture

The Model Context Protocol (MCP) provides a standardized interface for large language models (LLMs) like Claude to interact with external tools and services. It creates a unified API layer that simplifies integration and enables consistent interaction patterns.

Key components:
- **Tool Registry**: Defines available tools and their parameters
- **Function Calling**: Standardized format for LLM to invoke tools
- **Context Management**: Maintains state between interactions
- **Response Handling**: Structured format for tool responses

### Salesforce Integration Architecture

This implementation connects Claude to Salesforce through multiple API endpoints:
- **REST API**: For object CRUD operations and queries
- **SOAP API**: For metadata listing and deployment
- **Metadata API**: For creating and updating metadata components
- **Tooling API**: For faster metadata operations on individual components
- **Bulk API**: For efficient data operations

### Security Considerations

- **Authentication**: Credentials are provided directly in chat for demo purposes, but production implementations should use secure credential storage
- **Session Management**: Server maintains session tokens securely
- **Error Handling**: Prevents exposure of sensitive information in error messages
- **Input Validation**: Sanitizes inputs to prevent injection attacks

## MCP Tools Implementation Details

### Connection Tools

#### connect_to_salesforce
- **Implementation**: Uses simple-salesforce library with multiple authentication methods
- **Error Handling**: Attempts both separate token and password+token concatenation methods
- **Verification**: Performs test query to confirm connection validity
- **Performance**: Lightweight connection with minimal overhead

#### get_salesforce_connection_status
- **Implementation**: Executes a simple query against Organization object
- **Session Validation**: Verifies session token is still valid
- **Error Detection**: Identifies expired sessions before operations fail

### Metadata Tools

#### describe_metadata_types
- **Implementation**: Uses REST Metadata API's describe endpoint
- **Fallback Mechanism**: Provides predefined common types if API call fails
- **Response Processing**: Filters and formats metadata type information for clarity

#### list_metadata
- **Implementation**: Uses SOAP Metadata API's listMetadata operation
- **XML Processing**: Parses complex SOAP responses into structured data
- **Folder Support**: Handles folder-based metadata types properly

#### describe_object
- **Implementation**: Uses REST API's describe endpoint for object schema
- **Response Handling**: Returns complete object structure with all fields and relationships
- **Performance**: Caches results for frequently accessed objects

#### create_custom_field
- **Implementation**: Uses Tooling API for faster field creation
- **Field Type Mapping**: Handles conversion between common names and Salesforce API types
- **Metadata Construction**: Builds appropriate metadata structure based on field type
- **Validation**: Ensures field names meet Salesforce naming conventions

#### create_custom_object
- **Implementation**: Uses Metadata API for reliable object creation
- **XML Generation**: Constructs proper XML structure for object definition
- **Deployment Tracking**: Returns deployment ID for status monitoring
- **Naming Convention**: Enforces proper API naming rules

#### create_data_model
- **Implementation**: Uses Metadata API's deploy operation for transactional deployment
- **Relationship Handling**: Properly configures lookup and master-detail relationships
- **Error Recovery**: Provides detailed error information for failed deployments
- **Validation**: Verifies model specification before deployment

### SOQL Tools

#### generate_soql_query
- **Implementation**: Constructs SOQL strings with proper escaping
- **Auto-Field Selection**: Can automatically select fields if not specified
- **Injection Prevention**: Sanitizes inputs to prevent SOQL injection
- **Query Validation**: Ensures generated queries follow SOQL syntax rules

#### execute_soql_query
- **Implementation**: Uses REST API's query endpoint
- **Result Processing**: Removes Salesforce metadata from results for cleaner output
- **Pagination Handling**: Identifies when results are paginated
- **Error Handling**: Provides clear error messages for invalid queries

### Test Data Tools

#### generate_test_data
- **Implementation**: Uses Bulk API for efficient data insertion
- **Schema Analysis**: Determines required fields and their constraints
- **Value Generation**: Creates appropriate values based on field types
- **Themed Data**: Supports industry-specific or contextual data generation
- **Bulk Processing**: Handles errors individually without failing entire batch

### Utility Tools

#### get_org_limits
- **Implementation**: Uses REST API's limits endpoint
- **Response Processing**: Simplifies complex limits response for readability
- **Performance**: Lightweight operation with minimal API consumption

#### search_salesforce_documentation
- **Implementation**: Uses Google Custom Search API with Salesforce-specific scope
- **Result Formatting**: Structures search results with titles, links, and snippets
- **Error Handling**: Provides clear messages for API errors or rate limiting
- **Authentication**: Requires Google API key and Custom Search Engine ID

## Performance Considerations

- **API Efficiency**: Uses appropriate API for each operation type
- **Bulk Operations**: Batches requests when possible to reduce API calls
- **Error Recovery**: Implements retry logic for transient failures
- **Logging**: Comprehensive logging for troubleshooting and optimization

## Scalability Features

- **Connection Pooling**: Reuses connections when possible
- **Batch Processing**: Handles large datasets through batching
- **Parallel Execution**: Supports concurrent operations where appropriate
- **Resource Management**: Monitors and respects API limits

## Integration with Claude

The MCP Server exposes tools that Claude can discover and use through the Model Context Protocol. Claude can:
1. Understand user requests in natural language
2. Select the appropriate tool based on intent
3. Format parameters correctly for the tool
4. Process and explain tool responses
5. Handle errors and suggest alternatives

This creates a seamless experience where users can interact with Salesforce through natural conversation with Claude, without needing to learn complex APIs or query languages.
